/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_main.c -- client main loop

#include "quakedef.h"
extern int in_impulse;
// we need to declare some mouse variables here, because the menu system
// references them even when on a unix system.

// these two are not intended to be set directly
cvar_t	cl_name = {"_cl_name", "player", true};
cvar_t	cl_color = {"_cl_color", "0", true};

cvar_t	cl_shownet = {"cl_shownet","0"};	// can be 0, 1, or 2
cvar_t	cl_nolerp = {"cl_nolerp","0"};

cvar_t	lookspring = {"lookspring","0", true};
cvar_t	lookstrafe = {"lookstrafe","0", true};
cvar_t	sensitivity = {"sensitivity","3", true};

cvar_t	m_pitch = {"m_pitch","0.022", true};
cvar_t	m_yaw = {"m_yaw","0.022", true};
cvar_t	m_forward = {"m_forward","1", true};
cvar_t	m_side = {"m_side","0.8", true};

//PROQUAKE ADDITIONS _START
// JPG - added these for %r formatting
cvar_t	pq_needrl = {"pq_needrl", "I need RL", true};
cvar_t	pq_haverl = {"pq_haverl", "I have RL", true};
cvar_t	pq_needrox = {"pq_needrox", "I need rockets", true};

// JPG - added these for %p formatting
cvar_t	pq_quad = {"pq_quad", "quad", true};
cvar_t	pq_pent = {"pq_pent", "pent", true};
cvar_t	pq_ring = {"pq_ring", "eyes", true};

// JPG 3.00 - added these for %w formatting
cvar_t	pq_weapons = {"pq_weapons", "SSG:NG:SNG:GL:RL:LG", true};
cvar_t	pq_noweapons = {"pq_noweapons", "no weapons", true};

// JPG 1.05 - translate +jump to +moveup under water
cvar_t	pq_moveup = {"pq_moveup", "0", true};

// JPG 3.00 - added this by request
cvar_t	pq_smoothcam = {"pq_smoothcam", "1", true};
//PROQUAKE ADDITIONS _END

// by joe
cvar_t	cl_truelightning	= {"cl_truelightning", "0"};
cvar_t	cl_sbar				= {"cl_sbar", "0", true};
cvar_t	cl_rocket2grenade	= {"cl_r2g", "0"};
cvar_t	cl_mapname			= {"mapname", ""};
cvar_t	cl_muzzleflash		= {"cl_muzzleflash", "1"};
cvar_t	r_glowlg			= {"r_glowlg", "1",true};//R00k
cvar_t	r_powerupglow		= {"r_powerupglow", "2",true};//0=off, 1= white, 2=colored
cvar_t	cl_explosiontype		= {"cl_explosiontype", "0",true};
cvar_t	r_explosionlight	= {"r_explosionlight", "1",true};
cvar_t	r_rocketlight		= {"r_rocketlight", "1",true};
#ifdef GLQUAKE
cvar_t	r_explosionlightcolor	= {"r_explosionlightcolor", "0"};
cvar_t	r_rocketlightcolor	= {"r_rocketlightcolor", "0"};
#endif
cvar_t	r_rockettrail		= {"r_rockettrail", "1",true};
cvar_t	r_grenadetrail		= {"r_grenadetrail", "1",true};

cvar_t	cl_bobbing			= {"cl_bobbing", "1",true};
cvar_t	cl_demospeed		= {"cl_demospeed", "1"};
cvar_t	cl_deadbodyfilter	= {"cl_deadbodyfilter", "0",true};
cvar_t	cl_gibfilter		= {"cl_gibfilter", "0",true};
cvar_t	cl_maxfps			= {"cl_maxfps", "72", true, true};
cvar_t	cl_advancedcompletion	= {"cl_advancedcompletion", "1"};
client_static_t		cls;
client_state_t		cl;
// FIXME: put these on hunk?
efrag_t			cl_efrags[MAX_EFRAGS];
entity_t		cl_entities[MAX_EDICTS];
entity_t		cl_static_entities[MAX_STATIC_ENTITIES];
lightstyle_t		cl_lightstyle[MAX_LIGHTSTYLES];
dlight_t		cl_dlights[MAX_DLIGHTS];

int			cl_numvisedicts;
entity_t		*cl_visedicts[MAX_VISEDICTS];

modelindex_t		cl_modelindex[NUM_MODELINDEX];
char			*cl_modelnames[NUM_MODELINDEX];

tagentity_t		q3player_body, q3player_head;
unsigned CheckModel (char *mdl);

//----------
//Bestweapon
//----------
void Bestweapon_f (void)
{
	int i=1;
	if (Cmd_Argc() == 1)
	{
		Con_Printf ("Use \"BestWeapon X Y Z...\"\n");
		Con_Printf ("where X is the first option and Y the second one\n");
		return;
	}

	for (i=1; i<=Cmd_Argc(); i++)
	{
		switch(atoi(Cmd_Argv(i)))
		{
			case 8:
			if (cl.items & IT_LIGHTNING && cl.stats[STAT_CELLS] > 0)
			{
				in_impulse = 8;
				i = Cmd_Argc() + 1;
			}
			break;
			case 7:
			if (cl.items & IT_ROCKET_LAUNCHER && cl.stats[STAT_ROCKETS] > 0)
			{
				in_impulse = 7;
				i = Cmd_Argc() + 1;
			}
			break;

			case 6:
			if (cl.items & IT_GRENADE_LAUNCHER && cl.stats[STAT_ROCKETS] > 0)
			{
				in_impulse = 6;
				i = Cmd_Argc() + 1;
			}
			break;
			case 5:
			if (cl.items & IT_SUPER_NAILGUN && cl.stats[STAT_NAILS] > 0)
			{
				in_impulse = 5;
				i = Cmd_Argc() + 1;
			}
			break;
			case 4:
			if (cl.items & IT_NAILGUN && cl.stats[STAT_NAILS] > 0)
			{
				in_impulse = 4;
				i = Cmd_Argc() + 1;
			}
			break;
			case 3:
			if (cl.items & IT_SUPER_SHOTGUN && cl.stats[STAT_SHELLS]> 0)
			{
				in_impulse = 3;
				i = Cmd_Argc() + 1;
			}
			break;
			case 2:
			if (cl.items & IT_SHOTGUN && cl.stats[STAT_SHELLS] > 0)
			{
				in_impulse = 2;
				i = Cmd_Argc() + 1;
			}
			break;
			case 1:
			in_impulse = 1;
			i = Cmd_Argc() + 1;
			break;
		}
	}
}
/*
=====================
CL_ClearState
=====================
*/
#ifdef GLQUAKE
extern int	coronas;
#endif
void CL_ClearState (void)
{
	int	i;

	if (!sv.active)
	Host_ClearMemory ();

	CL_ClearTEnts ();
	
	// wipe the entire cl structure
	memset (&cl, 0, sizeof(cl));

	SZ_Clear (&cls.message);

	// clear other arrays
	memset (cl_efrags, 0, sizeof(cl_efrags));
	memset (cl_entities, 0, sizeof(cl_entities));
	memset (cl_dlights, 0, sizeof(cl_dlights));
	memset (cl_lightstyle, 0, sizeof(cl_lightstyle));
	memset (cl_temp_entities, 0, sizeof(cl_temp_entities));
	memset (&cl_beams, 0, sizeof(cl_beams));
#ifdef GLQUAKE
	coronas = 0;
#endif
	memset (cl_static_entities, 0, sizeof(cl_static_entities)); 


	// allocate the efrags and chain together into a free list
	cl.free_efrags = cl_efrags;
	for (i=0 ; i<MAX_EFRAGS-1 ; i++)
	cl.free_efrags[i].entnext = &cl.free_efrags[i+1];
	cl.free_efrags[i].entnext = NULL;

	#ifdef GLQUAKE
	#ifdef WITH_FMOD
	if (nehahra)
	SHOWLMP_clear ();
	#endif
	#endif
}

#ifdef GLQUAKE
#ifdef WITH_FMOD
// nehahra supported
void Neh_ResetSFX (void)
{
	int	i;

	if (num_sfxorig == 0)
	num_sfxorig = num_sfx;

	num_sfx = num_sfxorig;
	Con_DPrintf ("Current SFX: %d\n", num_sfx);

	for (i=num_sfx+1 ; i<MAX_SFX ; i++)
	{
		strcpy (known_sfx[i].name, "dfw3t23EWG#@T#@");
		if (known_sfx[i].cache.data)
		Cache_Free (&known_sfx[i].cache);
	}
}
#endif
#endif

/*
=====================
CL_Disconnect

Sends a disconnect message to the server
This is also called on Host_Error, so it shouldn't cause any errors
=====================
*/
void CL_Disconnect (void)
{
	// stop sounds (especially looping!)
	S_StopAllSounds (true);

	#ifdef GLQUAKE
	#ifdef WITH_FMOD
	if (nehahra)
		FMOD_Stop_f ();
	#endif
	#endif

	// if running a local server, shut it down
	if (cls.demoplayback)
	{
		CL_StopPlayback ();
	}
	else if (cls.state == ca_connected)
	{
		if (cls.demorecording)
		CL_Stop_f ();

		Con_DPrintf ("Sending clc_disconnect\n");
		SZ_Clear (&cls.message);
		MSG_WriteByte (&cls.message, clc_disconnect);
		NET_SendUnreliableMessage (cls.netcon, &cls.message);
		SZ_Clear (&cls.message);
		NET_Close (cls.netcon);

		cls.state = ca_disconnected;
		if (sv.active)
		Host_ShutdownServer (false);
	}

	cls.demoplayback = cls.timedemo = false;
	cls.signon = 0;

	#ifdef GLQUAKE
	#ifdef WITH_FMOD
	if (nehahra)
	Neh_ResetSFX ();
	#endif
	#endif
}

void CL_Disconnect_f (void)
{
	CL_Disconnect ();
	if (sv.active)
	Host_ShutdownServer (false);
}


/*
=====================
CL_EstablishConnection

Host should be either "local" or a net address to be passed on
=====================
*/
void CL_EstablishConnection (char *host)
{
	if (cls.state == ca_dedicated)
	return;

	#ifdef GLQUAKE
	#ifdef WITH_FMOD
	if (nehahra)
	num_sfxorig = num_sfx;
	#endif
	#endif

	if (cls.demoplayback)
	return;

	CL_Disconnect ();

	cls.netcon = NET_Connect (host);

	if (!cls.netcon)
	{		
		Con_Printf ("\nsyntax: connect server:port (port is optional)\n");//r00k added
		if (net_hostport != 26000)
			Con_Printf ("\nTry using port 26000\n");//r00k added
		Host_Error ("CL_Connect: connect failed");				
	}

	Con_DPrintf ("CL_EstablishConnection: connected to %s\n", host);

	if (cls.netcon->mod == MOD_PROQUAKE)
	Con_Printf("%c%cConnected to a ProQuake server%c\n", 1, 29, 31);

	cls.demonum = -1;			// not in the demo loop now
	cls.state = ca_connected;
	cls.signon = 0;				// need all the signon messages before playing

	MSG_WriteByte (&cls.message, clc_nop);	// joe: fix for NAT from ProQuake
}

/*
=====================
CL_SignonReply

An svc_signonnum has been received, perform a client side setup
=====================
*/

void CL_SignonReply (void)
{
	Con_DPrintf ("CL_SignonReply: %i\n", cls.signon);

	switch (cls.signon)
	{
		case 1:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "prespawn");
		break;

		case 2:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("name \"%s\"\n", cl_name.string));

		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("color %i %i\n", ((int)cl_color.value) >> 4, ((int)cl_color.value) & 15));

		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("spawn %s", cls.spawnparms));
		break;

		case 3:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "begin");
		Cache_Report ();		// print remaining memory
		break;

		case 4:
		SCR_EndLoadingPlaque ();	// allow normal screen updates
		if ((cl_autodemo.value == 1) && (!cls.demoplayback))
			Cmd_ExecuteString ("record\n", src_command);
		break;
	}
}

/*
=====================
CL_NextDemo

Called to play the next demo in the demo loop
=====================
*/
int CL_NextDemo (void)
{
	char	str[128];

	if (cls.demonum == -1)
	return 0;		// don't play demos

	SCR_BeginLoadingPlaque ();

	if (!cls.demos[cls.demonum][0] || cls.demonum == MAX_DEMOS)
	{
		cls.demonum = 0;
		if (!cls.demos[cls.demonum][0])
		{
			Con_Printf ("No demos listed with startdemos\n");
			cls.demonum = -1;

			return 0;
		}
	}

	Q_snprintfz (str, sizeof(str), "playdemo %s\n", cls.demos[cls.demonum]);
	Cbuf_InsertText (str);
	cls.demonum++;

	return 1;
}

/*
==============
CL_PrintEntities_f
==============
*/
void CL_PrintEntities_f (void)
{
	entity_t	*ent;
	int		i;

	for (i = 0, ent = cl_entities ; i < cl.num_entities ; i++, ent++)
	{
		Con_Printf ("%3i:", i);
		if (!ent->model)
		{
			Con_Printf ("EMPTY\n");
			continue;
		}
		Con_Printf ("%s:%2i  (%5.1f,%5.1f,%5.1f) [%5.1f %5.1f %5.1f]\n", ent->model->name, ent->frame, ent->origin[0], ent->origin[1], ent->origin[2], ent->angles[0], ent->angles[1], ent->angles[2]);
	}
}

dlighttype_t SetDlightColor (float f, dlighttype_t def, qboolean random)
{
	dlighttype_t colors[NUM_DLIGHTTYPES - 4] = {lt_red, lt_blue, lt_redblue, lt_green, lt_white};

	if ((int) f == 1)
		return lt_red;
	else if ((int) f == 2)
		return lt_blue;
	else if ((int) f == 3) 
		return lt_redblue;
	else if ((int) f == 4)
		return lt_green;
	else if ((int) f == 5)
		return lt_white;
	else if (((int) f == NUM_DLIGHTTYPES - 5) && random)
		return colors[rand() % (NUM_DLIGHTTYPES - 4)];
	else
		return def;
}

/*
===============
CL_AllocDlight
===============
*/
dlight_t *CL_AllocDlight (int key)
{
	int		i;
	dlight_t	*dl;

	// first look for an exact key match
	if (key)
	{
		dl = cl_dlights;
		for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
		{
			if (dl->key == key)
			{
				memset (dl, 0, sizeof(*dl));
				dl->key = key;
				return dl;
			}
		}
	}

	// then look for anything else
	dl = cl_dlights;
	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time)
		{
			memset (dl, 0, sizeof(*dl));
			dl->key = key;
			return dl;
		}
	}

	dl = &cl_dlights[0];
	memset (dl, 0, sizeof(*dl));
	dl->key = key;
	return dl;
}


void CL_NewDlight (int key, vec3_t origin, float radius, float time, int type)
{
	dlight_t	*dl;

	dl = CL_AllocDlight (key);
	VectorCopy (origin, dl->origin);
	dl->radius = radius;
	dl->die = cl.time + time;
	#ifdef GLQUAKE
	dl->type = type;
	#endif
}

/*
===============
CL_DecayLights
===============
*/
void CL_DecayLights (void)
{
	int		i;
	dlight_t	*dl;

	dl = cl_dlights;
	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time || !dl->radius)
		continue;

		dl->radius -= host_frametime * dl->decay;
		if (dl->radius < 0)
		dl->radius = 0;
	}
}

/*
===============
CL_LerpPoint

Determines the fraction between the last two messages that the objects
should be put at.
===============
*/

float CL_LerpPoint (void)
{
	float	f, frac;

	f = cl.mtime[0] - cl.mtime[1];

	if (!f || cl_nolerp.value || cls.timedemo || sv.active)
	{
		cl.ctime = cl.mtime[0];
		// joe: I don't remember why I've added this condition originally...
		// lets keep cl.time's original id1 src values, so we keep up the pace
		// if demorecording with this exe turns to be serious
		//	if (!cl_demorewind.value || !cls.demoplayback)
		cl.time = cl.mtime[0];
		return 1;
	}

	if (f > 0.1)
	{	// dropped packet, or start of demo
		cl.mtime[1] = cl.mtime[0] - 0.1;
		f = 0.1;
	}
	frac = (cl.ctime - cl.mtime[1]) / f;
	if (frac < 0)
	{
		if (frac < -0.01)
		{
			cl.ctime = cl.mtime[1];
			//	if (!cl_demorewind.value || !cls.demoplayback)
			cl.time = cl.mtime[1];
		}
		frac = 0;
	}
	else if (frac > 1)
	{
		if (frac > 1.01)
		{
			cl.ctime = cl.mtime[0];
			//	if (!cl_demorewind.value || !cls.demoplayback)
			cl.time = cl.mtime[0];
		}
		frac = 1;
	}

	return frac;
}

qboolean Model_isDead (int modelindex, int frame)
{
	if (cl_deadbodyfilter.value == 2)
	{
		if ((modelindex == cl_modelindex[mi_fish] && frame >= 18 && frame <= 38) ||
		    (modelindex == cl_modelindex[mi_dog] && frame >= 8 && frame <= 25) ||
		    (modelindex == cl_modelindex[mi_soldier] && frame >= 8 && frame <= 28) ||
		    (modelindex == cl_modelindex[mi_enforcer] && frame >= 41 && frame <= 65) ||
		    (modelindex == cl_modelindex[mi_knight] && frame >= 76 && frame <= 96) ||
		    (modelindex == cl_modelindex[mi_hknight] && frame >= 42 && frame <= 62) ||
		    (modelindex == cl_modelindex[mi_scrag] && frame >= 46 && frame <= 53) ||
		    (modelindex == cl_modelindex[mi_ogre] && frame >= 112 && frame <= 135) ||
		    (modelindex == cl_modelindex[mi_fiend] && frame >= 45 && frame <= 53) ||
		    (modelindex == cl_modelindex[mi_vore] && frame >= 16 && frame <= 22) ||
		    (modelindex == cl_modelindex[mi_shambler] && frame >= 83 && frame <= 93) ||
		    (modelindex == cl_modelindex[mi_player] && frame >= 41 && frame <= 102))
			return true;
	}
	else
	{
		if ((modelindex == cl_modelindex[mi_fish] && frame == 38) ||
		    (modelindex == cl_modelindex[mi_dog] && (frame == 16 || frame == 25)) ||
		    (modelindex == cl_modelindex[mi_soldier] && (frame == 17 || frame == 28)) ||
		    (modelindex == cl_modelindex[mi_enforcer] && (frame == 54 || frame == 65)) ||
		    (modelindex == cl_modelindex[mi_knight] && (frame == 85 || frame == 96)) ||
		    (modelindex == cl_modelindex[mi_hknight] && (frame == 53 || frame == 62)) ||
		    (modelindex == cl_modelindex[mi_scrag] && frame == 53) ||
		    (modelindex == cl_modelindex[mi_ogre] && (frame == 125 || frame == 135)) ||
		    (modelindex == cl_modelindex[mi_fiend] && frame == 53) ||
		    (modelindex == cl_modelindex[mi_vore] && frame == 22) ||
		    (modelindex == cl_modelindex[mi_shambler] && frame == 93) ||
		    (modelindex == cl_modelindex[mi_player] && (frame == 49 || frame == 60 || frame == 69 || 
			frame == 84 || frame == 93 || frame == 102)))
			return true;
	}

	return false;
}

qboolean Model_isHead (int modelindex)
{
	if (modelindex == cl_modelindex[mi_h_dog] || modelindex == cl_modelindex[mi_h_soldier] || 
	    modelindex == cl_modelindex[mi_h_enforcer] || modelindex == cl_modelindex[mi_h_knight] || 
	    modelindex == cl_modelindex[mi_h_hknight] || modelindex == cl_modelindex[mi_h_scrag] || 
	    modelindex == cl_modelindex[mi_h_ogre] || modelindex == cl_modelindex[mi_h_fiend] || 
	    modelindex == cl_modelindex[mi_h_vore] || modelindex == cl_modelindex[mi_h_shambler] || 
	    modelindex == cl_modelindex[mi_h_zombie] || modelindex == cl_modelindex[mi_h_player])
		return true;

	return false;
}

extern cvar_t pq_timer; // JPG - need this for CL_RelinkEntities

/*
===============
CL_RelinkEntities
===============
*/
void QMB_MuzzleFlash (vec3_t org);
void QMB_MuzzleFlashLG (vec3_t org);
void QMB_AxeSwoosh(vec3_t org);

void CL_RelinkEntities (void)
{
	entity_t	*ent;
	int			i, j;
	float		frac, f, d, bobjrotate;
	vec3_t		delta, oldorg;
	dlight_t	*dl;
	model_t		*model;
	dlighttype_t dimlightcolor;

	// determine partial update time
	frac = CL_LerpPoint ();

	// JPG - check to see if we need to update the status bar
	if (pq_timer.value && (cl.time != cl.oldtime))
	Sbar_Changed();

	cl_numvisedicts = 0;

	// interpolate player info
	for (i=0 ; i<3 ; i++)
	cl.velocity[i] = cl.mvelocity[1][i] + frac * (cl.mvelocity[0][i] - cl.mvelocity[1][i]);
	//PROQUAKE ADDITION --START
	if (cls.demoplayback || ((last_angle_time > host_time && !(in_attack.state & 3)) && pq_smoothcam.value)) // JPG - check for last_angle_time for smooth chasecam!
	{
		// interpolate the angles
		for (j=0 ; j<3 ; j++)
		{
			d = cl.mviewangles[0][j] - cl.mviewangles[1][j];
			if (d > 180)
			d -= 360;
			else if (d < -180)
			d += 360;

			// JPG - I can't set cl.viewangles anymore since that messes up the demorecording.  So instead,
			// I'll set lerpangles (new variable), and view.c will use that instead.
			cl.lerpangles[j] = cl.mviewangles[1][j] + frac*d;
		}
	}
	else
	VectorCopy(cl.viewangles, cl.lerpangles);
	//PROQUAKE ADDITION --END

	bobjrotate = anglemod (100 * cl.time);

	// start on the entity after the world
	for (i = 1, ent = cl_entities + 1 ; i < cl.num_entities ; i++, ent++)
	{
		if (!ent->model)
		{	// empty slot
			if (ent->forcelink)
			R_RemoveEfrags (ent);	// just became empty
			continue;
		}

		// if the object wasn't included in the last packet, remove it
		if (ent->msgtime != cl.mtime[0])
		{
			ent->model = NULL;

			// fenix@io.com: model transform interpolation
			ent->frame_start_time = ent->translate_start_time = ent->rotate_start_time = 0;

			continue;
		}

		VectorCopy (ent->origin, oldorg);
		
		if (ent->forcelink)
		{	// the entity was not updated in the last message so move to the final spot
			VectorCopy (ent->msg_origins[0], ent->origin);
			VectorCopy (ent->msg_angles[0], ent->angles);
		}
		else
		{	// if the delta is large, assume a teleport and don't lerp
			f = frac;
			for (j=0 ; j<3 ; j++)
			{
				delta[j] = ent->msg_origins[0][j] - ent->msg_origins[1][j];
				if (delta[j] > 200 || delta[j] < -200)
				f = 1;		// assume a teleportation, not a motion
			}

			if (f >= 1)
			{
				ent->translate_start_time = 0;
				ent->rotate_start_time    = 0;
			}

			// interpolate the origin and angles
			for (j=0 ; j<3 ; j++)
			{
				ent->origin[j] = ent->msg_origins[1][j] + f*delta[j];

				d = ent->msg_angles[0][j] - ent->msg_angles[1][j];
				if (d > 180)
				d -= 360;
				else if (d < -180)
				d += 360;
				ent->angles[j] = ent->msg_angles[1][j] + f*d;
			}
		}

		if (cl_deadbodyfilter.value && ent->model->type == mod_alias && Model_isDead(ent->modelindex, ent->frame))
		continue;

		if (cl_gibfilter.value && ent->model->type == mod_alias && 
		    (ent->modelindex == cl_modelindex[mi_gib1] || ent->modelindex == cl_modelindex[mi_gib2] || 
		     ent->modelindex == cl_modelindex[mi_gib3] || Model_isHead(ent->modelindex)))
			continue;

		if (ent->modelindex == cl_modelindex[mi_explo1] || ent->modelindex == cl_modelindex[mi_explo2])
		{
			// software removal of sprites
			if (cl_explosiontype.value == 2 || cl_explosiontype.value == 3)
				continue;

#ifdef GLQUAKE
			if (qmb_initialized && gl_part_explosions.value)
			{
				R_RocketTrail (oldorg, ent->origin, LAVA_TRAIL);//R00k for flamethrower
				continue;
			}
#endif
		}

		if (!(model = cl.model_precache[ent->modelindex]))
		Host_Error ("CL_RelinkEntities: bad modelindex");

		if (ent->modelindex == cl_modelindex[mi_rocket] &&
		cl_rocket2grenade.value && cl_modelindex[mi_grenade] != -1)
		ent->model = cl.model_precache[cl_modelindex[mi_grenade]];

		// rotate binary objects locally
		if (ent->model->flags & EF_ROTATE)
		{
			if ((ent->model != cl.model_precache[cl_modelindex[mi_flag]]) && (ent->model != cl.model_precache[cl_modelindex[mi_w_s_key]]) && (ent->model != cl.model_precache[cl_modelindex[mi_w_g_key]])) //R00k dont rotate flags!
			{
				ent->angles[1] = bobjrotate;
				if (cl_bobbing.value)
				ent->origin[2] += sin(bobjrotate / 90 * M_PI) * 5 + 5;
			}
		}

		// EF_BRIGHTFIELD is not used by original progs
		if (ent->effects & EF_BRIGHTFIELD)
		{
			R_EntityParticles (ent);
		}

		if ((ent->effects & EF_MUZZLEFLASH) && cl_muzzleflash.value)
		{
			vec3_t	fv, rv, uv;
			vec3_t	smokeorg, smokeorg2;

			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin, dl->origin);
			dl->origin[2] += 16;
			AngleVectors (ent->angles, fv, rv, uv);
			VectorMA (dl->origin, 18, fv, dl->origin);
			dl->radius = 200 + (rand() & 31);
			dl->minlight = 32;
			dl->die = cl.time + 0.1;			
			#ifdef GLQUAKE
				if (ent->modelindex == cl_modelindex[mi_shambler] && qmb_initialized && gl_part_lightning.value)
					dl->type = lt_blue;
				else
				{
					if (ent->modelindex == cl_modelindex[mi_scrag] && qmb_initialized)
						dl->type = lt_green;
					else
						dl->type = lt_muzzleflash;
				}
			#endif

			// check for progs/player.mdl and if cl.stats[STAT_ACTIVEWEAPON] == whatever
			// and change the muzzle flash colour depending on the weapon
			if (i == cl.viewentity)
			{
				VectorCopy (cl_entities[cl.viewentity].origin, smokeorg);
				smokeorg[2] += cl.crouch;

				//VectorCopy (ent->origin, smokeorg);

				smokeorg[2] += 14;	// move it up

				// adjust for current view angles
				
				AngleVectors (cl.viewangles, fv, rv, uv);

				// forward vector
				VectorMA (smokeorg, 18, fv, smokeorg);

				if ((cl.stats[STAT_ACTIVEWEAPON] == IT_NAILGUN)||(cl.stats[STAT_ACTIVEWEAPON] == IT_SUPER_SHOTGUN))
				{
					VectorMA (smokeorg, -3, rv, smokeorg);
					VectorMA (smokeorg, 6, rv, smokeorg2);
					QMB_MuzzleFlash (smokeorg);					
					QMB_MuzzleFlash (smokeorg2);
				}
				else
				{
					if ((cl.stats[STAT_ACTIVEWEAPON] == IT_SUPER_NAILGUN)||(cl.stats[STAT_ACTIVEWEAPON] == IT_SHOTGUN))
					{
						QMB_MuzzleFlash (smokeorg);
					}
					else
					{
						if (cl.stats[STAT_ACTIVEWEAPON] == IT_LIGHTNING)
						{
							QMB_MuzzleFlashLG(smokeorg);
						}						
					}
				}
			}
		}
		/*
		if (ent->modelindex != cl_modelindex[i_eyes] &&
		((ent->modelindex != cl_modelindex[i_player] && ent->model->modhint != MOD_PLAYER &&
		ent->modelindex != cl_modelindex[i_h_player]) || (r_powerupglow.value && cl.stats[STAT_HEALTH] > 0)))
		*/
		if (r_powerupglow.value)
		{
			dimlightcolor = lt_default;

			if (r_powerupglow.value == 2)
			{
				dimlightcolor = SetDlightColor(lt_white, lt_default, false);//Set glow Full Bright!			
			}

			if ((ent->effects & (EF_BLUE | EF_RED)) == (EF_BLUE | EF_RED))
			{
				if (r_powerupglow.value == 1)
				{
					CL_NewDlight (i, ent->origin, 200, 0.1, lt_redblue);
					CL_NewDlight (i+1, ent->origin, 80, 0.1, dimlightcolor);
				}
				else
					CL_NewDlight (i, ent->origin, 200, 0.1, dimlightcolor);
			}
			else
			{
				if (ent->effects & EF_BLUE)
				{
					if (r_powerupglow.value == 1)
					{
						CL_NewDlight (i, ent->origin, 200, 0.1, lt_blue);
						CL_NewDlight (i+1, ent->origin, 80, 0.1, dimlightcolor);
					}
					else
					CL_NewDlight (i, ent->origin, 200, 0.1, dimlightcolor);
				}			
				else 
				{
					if (ent->effects & EF_RED)
					{
						if (r_powerupglow.value == 1)
						{
							CL_NewDlight (i, ent->origin, 200, 0.1, lt_red);
							CL_NewDlight (i+1, ent->origin, 80, 0.1, dimlightcolor);
						}
						else
							CL_NewDlight (i, ent->origin, 200, 0.1, dimlightcolor);
					}					
					else
					{
						if (ent->effects & EF_BRIGHTLIGHT)
						{
							vec3_t	tmp;
							VectorCopy (ent->origin, tmp);
							tmp[2] += 16;
							CL_NewDlight (i, tmp, 400 + (rand() & 31), 0.1, dimlightcolor);
						}
						else
						{			
							if (ent->effects & EF_DIMLIGHT)
							{
								CL_NewDlight (i, ent->origin, 200, 0.1, dimlightcolor);
							}							
						}
					}
				}
			}
		}
/*
		if (!strcmp(ent->model->name, "progs/s_light.spr"))
		{
			R_RocketTrail (oldorg, ent->origin, LAVA_TRAIL);
		}
*/
		if (!strcmp(ent->model->name, "progs/flame2.mdl"))			
		{	
			#ifdef GLQUAKE
			if (qmb_initialized && gl_part_flames.value)
			{
				QMB_BigTorchFlame (ent->origin);					
				continue;
			}
			#endif
		}

		//R00k added for nailtrails
		if ((!strcmp(ent->model->name, "progs/s_spike.mdl"))||(!strcmp(ent->model->name, "progs/spike.mdl")))
		{	
			#ifdef GLQUAKE
			if (qmb_initialized && gl_part_trails.value)
			{
				R_RocketTrail (oldorg, ent->origin, NAIL_TRAIL);
			}
			#endif
		}
						
		if (model->flags)
		{
			if (model->flags & EF_GIB)
			R_RocketTrail (oldorg, ent->origin, BLOOD_TRAIL);
			else if (model->flags & EF_ZOMGIB)
			R_RocketTrail (oldorg, ent->origin, SLIGHT_BLOOD_TRAIL);
			else if (model->flags & EF_TRACER)
			R_RocketTrail (oldorg, ent->origin, TRACER1_TRAIL);
			else if (model->flags & EF_TRACER2)
			R_RocketTrail (oldorg, ent->origin, TRACER2_TRAIL);
			else if (model->flags & EF_ROCKET)
			{
				if (model->modhint == MOD_LAVABALL)
				{
					R_RocketTrail (oldorg, ent->origin, LAVA_TRAIL);

					dl = CL_AllocDlight (i);
					VectorCopy (ent->origin, dl->origin);
					dl->radius = 100 * (1 + bound(0, r_rocketlight.value, 1));
					dl->die = cl.time + 0.1;
					#ifdef GLQUAKE
					dl->type = lt_rocket;
					#endif
				}
				else
				{
					if (r_rockettrail.value == 2)
					R_RocketTrail (oldorg, ent->origin, GRENADE_TRAIL);
					else if (r_rockettrail.value == 3)
					R_RocketTrail (oldorg, ent->origin, ALT_ROCKET_TRAIL);
					else if (r_rockettrail.value)
					R_RocketTrail (oldorg, ent->origin, ROCKET_TRAIL);

					if (r_rocketlight.value)
					{
						dl = CL_AllocDlight (i);
						VectorCopy (ent->origin, dl->origin);
						dl->radius = 100 * (1 + bound(0, r_rocketlight.value, 1));
						dl->die = cl.time + 0.1;
						#ifdef GLQUAKE
						dl->type = SetDlightColor (r_rocketlightcolor.value, lt_rocket, false);
						#endif
					}
				}
			}
			else if ((model->flags & EF_GRENADE) && r_grenadetrail.value)
			{
				// Nehahra dem compatibility
				if (ent->transparency == -1)
				{
					if (cl.time >= ent->smokepuff_time)
					{
						R_RocketTrail (oldorg, ent->origin, NEHAHRA_SMOKE);
						ent->smokepuff_time = cl.time + 0.14;
					}
				}
				else
				{
					R_RocketTrail (oldorg, ent->origin, GRENADE_TRAIL);
				}
			}
			else 
			if (model->flags & EF_TRACER3)
				R_RocketTrail (oldorg, ent->origin, VOOR_TRAIL);			
		}

		ent->forcelink = false;

		if (i == cl.viewentity && !chase_active.value)
		continue;

		// nehahra supported
		if (ent->effects & EF_NODRAW)
		continue;

		#ifdef GLQUAKE
		if (qmb_initialized)
		{
			if (ent->modelindex == cl_modelindex[mi_bubble])
			{
				if (!cl.paused && cl.oldtime != cl.time)
				QMB_StaticBubble (ent);
				continue;
			}
			else if (gl_part_lightning.value && ent->modelindex == cl_modelindex[mi_shambler] &&
			ent->frame >= 65 && ent->frame <= 68)
			{
				vec3_t	liteorg;

				VectorCopy(ent->origin, liteorg);
				liteorg[2] += 32;
				QMB_ShamblerCharge (liteorg);
			}			
		}
		#endif

		if (cl_numvisedicts < MAX_VISEDICTS)
		{
			cl_visedicts[cl_numvisedicts] = ent;
			cl_numvisedicts++;
		}
	}
};


/*
===============
CL_CalcCrouch

Smooth out stair step ups
===============
*/
void CL_CalcCrouch (void)
{
	qboolean	teleported;
	entity_t	*ent;
	static	vec3_t	oldorigin = {0, 0, 0};
	static	float	oldz = 0, extracrouch = 0, crouchspeed = 100;

	ent = &cl_entities[cl.viewentity];

	teleported = !VectorL2Compare(ent->origin, oldorigin, 48);
	VectorCopy(ent->origin, oldorigin);

	if (teleported)
	{
		// possibly teleported or respawned
		oldz = ent->origin[2];
		extracrouch = 0;
		crouchspeed = 100;
		cl.crouch = 0;
		return;
	}

	if (cl.onground && ent->origin[2] - oldz > 0)
	{
		if (ent->origin[2] - oldz > 20)
		{
			// if on steep stairs, increase speed
			if (crouchspeed < 160)
			{
				extracrouch = ent->origin[2] - oldz - host_frametime * 200 - 15;
				extracrouch = min(extracrouch, 5);
			}
			crouchspeed = 160;
		}

		oldz += host_frametime * crouchspeed;
		if (oldz > ent->origin[2])
		oldz = ent->origin[2];

		if (ent->origin[2] - oldz > 15 + extracrouch)
		oldz = ent->origin[2] - 15 - extracrouch;
		extracrouch -= host_frametime * 200;
		extracrouch = max(extracrouch, 0);

		cl.crouch = oldz - ent->origin[2];
	}
	else
	{
		// in air or moving down
		oldz = ent->origin[2];
		cl.crouch += host_frametime * 150;
		if (cl.crouch > 0)
		cl.crouch = 0;
		crouchspeed = 100;
		extracrouch = 0;
	}
}

/*
===============
CL_ReadFromServer

Read all incoming data from the server
===============
*/
int CL_ReadFromServer (void)
{
	int	ret;

	cl.oldtime = cl.ctime;
	cl.time += host_frametime;
	if (!cl_demorewind.value || !cls.demoplayback)	// by joe
	cl.ctime += host_frametime;
	else
	cl.ctime -= host_frametime;

	do {
		ret = CL_GetMessage ();
		if (ret == -1)
		Host_Error ("CL_ReadFromServer: lost server connection");
		if (!ret)
		break;

		cl.last_received_message = realtime;
		CL_ParseServerMessage ();
	} while (ret && cls.state == ca_connected);

	if (cl_shownet.value)
	Con_Printf ("\n");
	
	CL_RelinkEntities ();
	CL_CalcCrouch ();
	CL_UpdateTEnts ();

	// bring the links up to date
	return 0;
}

/*
=================
CL_SendCmd
=================
*/

void CL_SendCmd (void)
{
	usercmd_t	cmd;

	if (cls.state != ca_connected)
	return;

	if (cls.signon == SIGNONS)
	{
		CL_BaseMove (&cmd);
		IN_Move (&cmd);
		CL_SendMove (&cmd);
	}

	if (cls.demoplayback)
	{
		SZ_Clear (&cls.message);
		return;
	}

	// send the reliable message
	if (!cls.message.cursize)
	return;	// no message at all

	if (!NET_CanSendMessage(cls.netcon))
	{
		Con_DPrintf ("CL_WriteToServer: can't send\n");
		return;
	}

	if (NET_SendMessage(cls.netcon, &cls.message) == -1)
	Host_Error ("CL_WriteToServer: lost server connection");

	SZ_Clear (&cls.message);
}

// by joe
char *CL_MapName (void)
{
	return cl_mapname.string;
}

#ifdef GLQUAKE
void CL_Fog_f (void) 
{
	extern cvar_t gl_fogred, gl_foggreen, gl_fogblue, gl_fogenable;
	if (Cmd_Argc () == 1) {
		Con_Printf ("\"fog\" is \"%f %f %f\"\n", gl_fogred.value, gl_foggreen.value, gl_fogblue.value);
		return;
	}
	gl_fogenable.value = 1;
	gl_fogred.value    = atof(Cmd_Argv(1));
	gl_foggreen.value  = atof(Cmd_Argv(2));
	gl_fogblue.value   = atof(Cmd_Argv(3));
}
#endif

/*
=================
CL_Init
=================
*/
void CL_Init (void)
{
	SZ_Alloc (&cls.message, 1024);

	CL_InitInput ();
	CL_InitTEnts ();

	// register our commands
	Cvar_RegisterVariable (&cl_name);
	Cvar_RegisterVariable (&cl_color);
	Cvar_RegisterVariable (&cl_upspeed);
	Cvar_RegisterVariable (&cl_forwardspeed);
	Cvar_RegisterVariable (&cl_backspeed);
	Cvar_RegisterVariable (&cl_sidespeed);
	Cvar_RegisterVariable (&cl_movespeedkey);
	Cvar_RegisterVariable (&cl_yawspeed);
	Cvar_RegisterVariable (&cl_pitchspeed);
	Cvar_RegisterVariable (&cl_anglespeedkey);
	Cvar_RegisterVariable (&cl_shownet);
	Cvar_RegisterVariable (&cl_nolerp);
	Cvar_RegisterVariable (&lookspring);
	Cvar_RegisterVariable (&lookstrafe);
	Cvar_RegisterVariable (&sensitivity);

	Cvar_RegisterVariable (&m_pitch);
	Cvar_RegisterVariable (&m_yaw);
	Cvar_RegisterVariable (&m_forward);
	Cvar_RegisterVariable (&m_side);

	// by joe
	Cvar_RegisterVariable (&cl_truelightning);
	Cvar_RegisterVariable (&cl_sbar);
	Cvar_RegisterVariable (&cl_rocket2grenade);
	Cvar_RegisterVariable (&cl_demorewind);
//	Cvar_RegisterVariable (&cl_demoloop);
	Cvar_RegisterVariable (&cl_mapname);
	Cvar_RegisterVariable (&cl_muzzleflash);
	Cvar_RegisterVariable (&r_glowlg);
	Cvar_RegisterVariable (&r_powerupglow);
	Cvar_RegisterVariable (&cl_explosiontype);
	Cvar_RegisterVariable (&r_explosionlight);
	Cvar_RegisterVariable (&r_rocketlight);
	#ifdef GLQUAKE
	Cvar_RegisterVariable (&r_explosionlightcolor);
	Cvar_RegisterVariable (&r_rocketlightcolor);
	#endif
	Cvar_RegisterVariable (&r_rockettrail);
	Cvar_RegisterVariable (&r_grenadetrail);

	Cvar_RegisterVariable (&cl_bobbing);
	Cvar_RegisterVariable (&cl_demospeed);
	Cvar_RegisterVariable (&cl_deadbodyfilter);
	Cvar_RegisterVariable (&cl_gibfilter);
	Cvar_RegisterVariable (&cl_maxfps);
	Cvar_RegisterVariable (&cl_advancedcompletion);

	Cmd_AddCommand ("bestweapon", Bestweapon_f);

	// JPG - added these for %r formatting
	Cvar_RegisterVariable (&pq_needrl);
	Cvar_RegisterVariable (&pq_haverl);
	Cvar_RegisterVariable (&pq_needrox);

	// JPG - added these for %p formatting
	Cvar_RegisterVariable (&pq_quad);
	Cvar_RegisterVariable (&pq_pent);
	Cvar_RegisterVariable (&pq_ring);

	// JPG 3.00 - %w formatting
	Cvar_RegisterVariable (&pq_weapons);
	Cvar_RegisterVariable (&pq_noweapons);

	// JPG 1.05 - added this for +jump -> +moveup translation
	Cvar_RegisterVariable (&pq_moveup);

	// JPG 3.02 - added this by request
	Cvar_RegisterVariable (&pq_smoothcam);

	Cmd_AddCommand ("entities", CL_PrintEntities_f);
	Cmd_AddCommand ("disconnect", CL_Disconnect_f);
	Cmd_AddCommand ("record", CL_Record_f);
	Cmd_AddCommand ("stop", CL_Stop_f);
	Cmd_AddCommand ("playdemo", CL_PlayDemo_f);
	Cmd_AddCommand ("timedemo", CL_TimeDemo_f);
	#ifdef GLQUAKE
	Cmd_AddCommand ("fog",CL_Fog_f);//R00k
	#endif
}
